
TITLE ROM BIOS FOR MPX-16 SINGLE BOARD COMPUTER 
PAGE        60,132
;*
;*	DATE OF THIS REVISION:	8/23/83
;*
;----------------------------
; EQUATES
;----------------------------
CODESG      EQU    0FE00H       ; FOR 8K EPROM AT HIGH MEMORY
P8155_CMD   EQU    01C0H        ; 8155 COMMAND PORT
P8155_L     EQU    01C4H        ; 8155 LOW BYTE TIMER COUNT
P8155_H     EQU    01C5H        ; 8155 HIGH BYTE TIMER COUNT
P8255_A     EQU    01A0H        ; 8255 PORT A ADDRESS
P8255_B     EQU    01A1H        ; 8255 PORT B ADDRESS
P8255_C     EQU    01A2H        ; 8255 PORT C ADDRESS
P8255_D     EQU    01A3H        ; 8255 COMMAND PORT
P8259_M0    EQU    0140H        ; 8259 MASTER ADDRESS 0
P8259_M1    EQU    0141H        ; 8259 MASTER ADDRESS 1
P8259_S0    EQU    0120H        ; 8259 SLAVE ADDRESS 0
P8259_S1    EQU    0121H        ; 8259 SLAVE ADDRESS 1
P8253_0     EQU    01E0H        ; 8253 TIMER 1
P8253_1     EQU    01E1H        ; 8253 TIMER 2
P8253_2     EQU    01E2H        ; 8253 TIMER 3
P8253_CMD   EQU    01E3H        ; 8253 COMMAND PORT
P8251_CON   EQU    0161H        ; CON 8251 STATUS
P8251_AUX   EQU    0181H        ; AUX 8251 STATUS
P8237       EQU    0            ; 8237 BASE ADDRESS
P8272_STAT  EQU    20H          ; FDC STATUS REGISTER
P8272_DATA  EQU    21H          ; FDC DATA REGISTER
;
;	STATUS EQUATES FOR DISK I/O ROUTINES
;
TIME_OUT	EQU	080H	; ATTACHMENT FAILED TO RESPOND
BAD_SEEK	EQU	040H	; SEEK OPERATION FAILED
BAD_8272	EQU	020H	; FDC CONTROLLER CHIP FAILURE
BAD_CRC		EQU	010H	; CRC ERROR ON DISK READ
DMA_BOUNDARY	EQU	009H	; ATTEMPT TO DMA ACROSS 64K BOUNDARY
BAD_DMA		EQU	008H	; DMA OVERRUN ON OPERATION
RECORD_NOT_FND	EQU	004H	; REQUESTED SECTOR NOT FOUND
WRITE_PROTECT	EQU	003H	; WRITE ATTEMPTED ON PROTECTED DISK
BAD_ADDR_MARK	EQU	002H	; ADDRESS MARK NOT FOUND
BAD_CMD		EQU	001H	; BAD COMMAND SENT TO BIOS
PAGE_PORT	EQU	060H	; PAGE REGISTER FOR DMA, HI NIBBLE
MOTOR_PORT	EQU	0A0H	; MOTOR CONTROL, FDC RESET BIT
;
;	EXTERNAL PROCEDURES
;
EXTRN	PRINT_MESSAGE:FAR,KB_INIT:FAR,PRTSC_INT:FAR
EXTRN	KB_IO:FAR,KB_INT_HDLR:FAR,VID_INT_HDLR:FAR,VID_PARM_TABLE:FAR
;
;
;----------------------------
; INTERRUPT LOCATIONS
;----------------------------
ABS0       SEGMENT  AT 0
;
EXTRN	INT_0_VEC:BYTE,NMI_VEC:WORD,PRTSC_VEC:WORD,IBM_INT:WORD
EXTRN	DISK_POINTER:DWORD,VID_PARM_PTR:DWORD,GR_CHAR_EXT:DWORD
EXTRN	MPX_INT:WORD
;
ABS0	ENDS

;----------------------------
; INITIALIZATION STACK
;----------------------------
STACK      SEGMENT  AT 030H
           DW       128 DUP(?)
TOP_STACK  LABEL    WORD
STACK      ENDS

;----------------------------
; RESERVED DATA AREAS
;----------------------------
DATA       SEGMENT  AT 040H
;
EXTRN	EQUIP_FLAG:WORD,MEM_SIZ:WORD,TIMER_LSW:WORD,TIMER_MSW:WORD
EXTRN	TIMER_FLG:BYTE,RECAL_FLG:BYTE,MOTOR_STAT:BYTE,MOTOR_CNT:BYTE
EXTRN	DISK_STAT:BYTE,DISK_STAT_ARRAY:BYTE,NUMSEC:BYTE
EXTRN	DISK_INT_FLG:BYTE,BREAK_FLAG:BYTE
;
DATA         ENDS
;
;----------------------------
; ROM RESIDENT CODE   
;----------------------------
;
PUBLIC	START
;
CSEG	SEGMENT	COMMON

ASSUME    CS:CSEG,SS:STACK,ES:ABS0,DS:DATA

	ORG	01000H		; START AT FE00:1000
	DB	'MPX-16 BIOS COPR. MICROMINT 1983'

	DB	16 DUP(0FFH)		; RESERVE SOME SPACE
;
START	PROC
;
    CLI                            ; DISABLE INTERRUPTS

;----------------------------
; SEGMENT REGISTER INIT
;----------------------------

    MOV    AX,STACK               ; SETUP SS SEG REG
    MOV    SS,AX
    MOV    AX,OFFSET TOP_STACK    ; SETUP SP REG
    MOV    SP,AX
    MOV    AX,DATA                ; SETUP DS REG
    MOV    DS,AX
    MOV    AX,ABS0                ; SETUP ES REG
    MOV    ES,AX

;----------------------------
; HARDWARE INITIALIZATION
;----------------------------

    MOV    DX,P8255_D             ; 8255 COMMAND PORT
    MOV    AL,91H                 ; A + CL IN, B + CH OUT
    OUT    DX,AL                  ; INIT CHIP
    MOV    DX,P8255_C             ; DISABLE PARITY CHECK
    MOV    AL,0F0H
    OUT    DX,AL                  ; PARITY DISABLED 
    MOV    DX,P8155_CMD
    MOV    AL,040H
    OUT    DX,AL                  ; STOP RAM REFRESH TIMER
    MOV    DX,P8155_L
    MOV    AL,23H
    OUT    DX,AL                  ; LOAD LOW ORDER BYTE FOR TIMER
    MOV    DX,P8155_H
    MOV    AL,0C0H
    OUT    DX,AL                  ; LOAD HIGH ORDER BYTE FOR TIMER
    MOV    DX,P8155_CMD
    MOV    AL,0C0H
    OUT    DX,AL                  ; START RAM REFRESH TIMER

; DMA AND TIMER INITIALIZATION FOR MEMORY REFRESH

    MOV    AL,0                   ; MASTER CLEAR OF 8237
    OUT    P8237+0DH,AL
    MOV    AL,24H                 ; DISABLE 8237
    OUT    P8237+08H,AL
    MOV    AL,0                   ; SET ALL ADDRESSES TO 0
    OUT    P8237+0CH,AL           ; SET F/F
    OUT    P8237+00H,AL
    OUT    P8237+00H,AL
    OUT    P8237+02H,AL
    OUT    P8237+02H,AL
    OUT    P8237+04H,AL
    OUT    P8237+04H,AL
    OUT    P8237+06H,AL
    OUT    P8237+06H,AL
    OUT    040H,AL                ; CLEAR PAGE REGISTERS
    OUT    060H,AL
    OUT    080H,AL
    MOV    AL,0FFH                ; SET ALL COUNTS HIGH
    OUT    P8237+01H,AL
    OUT    P8237+01H,AL
    OUT    P8237+03H,AL
    OUT    P8237+03H,AL
    OUT    P8237+05H,AL
    OUT    P8237+05H,AL
    OUT    P8237+07H,AL
    OUT    P8237+07H,AL

    MOV    AL,58H                 ; SET MODE FOR CHANNEL 0
    OUT    P8237+0BH,AL
    MOV    AL,41H                 ; SET MODE FOR CHANNEL 1
    OUT    P8237+0BH,AL
    INC    AL                     ; SET MODE FOR CHANNEL 2
    OUT    P8237+0BH,AL
    INC    AL                     ; SET MODE FOR CHANNEL 3
    OUT    P8237+0BH,AL
    MOV    AL,0                   ; ENABLE DMA
    OUT    P8237+08H,AL
    OUT    P8237+0AH,AL

; INTERRUPT CONTROLLER CHIPS (MASTER AND SLAVE) INITIALIZATION

; INITIALIZE MASTER

    MOV    DX,P8259_M0            ; M0-MASTER, PORT 0
    MOV    AL,011H                ; EDGE-TRIGGER, CASCADE MODE, ICW4 NEEDED
    OUT    DX,AL                  ; ICW1
    MOV    DX,P8259_M1            ; M1-MASTER, PORT 1
    MOV    AL,040H                ; INTERRUPTS 40H-47H                       
    OUT    DX,AL                  ; ICW2
    MOV    AL,002H                ; IR1 HAS A SLAVE                        
    OUT    DX,AL                  ; ICW3
    MOV    AL,00DH                ; BUFFERED MODE/MASTER, NORMAL EOI, 8088 MODE
    OUT    DX,AL                  ; ICW4
    MOV    AL,0FCH                ; MASK INTERRUPTS (ALL EXCEPT RTI, SLV)
    OUT    DX,AL                  ; OCW1

; INITIALIZE SLAVE

    MOV    DX,P8259_S0            ; S0-SLAVE, PORT 0
    MOV    AL,011H                ; EDGE-TRIGGER, CASCADE MODE, ICW4 NEEDED
    OUT    DX,AL                  ; ICW1
    MOV    DX,P8259_S1            ; S1-SLAVE, PORT 1
    MOV    AL,048H                ; INTERRUPTS 48H-4FH                       
    OUT    DX,AL                  ; ICW2
    MOV    AL,001H                ; SLAVE ID = 1                 
    OUT    DX,AL                  ; ICW3
    MOV    AL,009H                ; BUFFERED MODE/SLAVE, NORMAL EOI, 8088 MODE
    OUT    DX,AL                  ; ICW4
    MOV    AL,0DFH                ; MASK INTERRUPTS (ALL EXCEPT DISK)
    OUT    DX,AL                  ; OCW1

 ; INITIALIZE 8253 TIMER CHIP FOR REAL TIME INTERRUPT 

    MOV    DX,P8253_CMD           ; COMMAND PORT TO 8253
    MOV    AL,0B6H                ; CONTROL WORD
    OUT    DX,AL
    DEC    DX                     ; COUNT ADDRESS FOR CHANNEL 3
	XOR	AX,AX		; 53.3 MS (18.75 INTERRUPTS PER SECOND)
    OUT    DX,AL                  ; LSB BYTE
    MOV    AL,AH
    OUT    DX,AL                  ; MSB BYTE
	XOR	AX,AX
	MOV	TIMER_LSW,AX	; SET TIME TO 0
	MOV	TIMER_MSW,AX
	MOV	TIMER_FLG,AL


;-----------------------------
; SIZE MEMORY
;-----------------------------

    PUSH   ES
    MOV    AX,1000H              ; ASSUME 64K
    MOV    ES,AX
	MOV	BX,64		; SIZE COUNTER
    MOV    SI,0
	MOV	DH,055H		; TEST PATTERN
MS1: MOV    ES:[SI],DH  ; WRITE TEST PATTERN
	MOV	DL,ES:[SI]
    CMP    DL,DH		; PATTERN STILL THERE?
    JNE    MS2
	ADD	BX,64		; ADD ANOTHER 64K TO SIZE COUNTER
    ADD    AX,1000H              ; NEXT 64K BLOCK
    CMP    AX,0B000H             ; 640K MAX
    JE     MS2
    MOV    ES,AX
    JMP    SHORT MS1
MS2: MOV    MEM_SIZ,BX
    POP    ES
    
;-----------------------------
; INITIALIZE VECTOR TABLE
;-----------------------------

; INITIALIZE NMI INTERRUPT 

    MOV    AX,OFFSET NMI_INT      ; INIT OFFSET OF NMI SERVICE ROUTINE
    MOV    ES:NMI_VEC,AX
    MOV    AX,CODESG              ; INIT SEGMENT OF NMI SERVICE ROUTINE
    MOV    ES:NMI_VEC+2,AX

;
; INITIALIZE PRINT SCREEN INTERRUPT
;
    MOV    AX,OFFSET PRTSC_INT   ; PRINT SCREEN INTERRUPT VECTOR
    MOV    ES:PRTSC_VEC,AX
    MOV    AX,CODESG
    MOV    ES:PRTSC_VEC+2,AX
;
; INITIALIZE SOFT INTERRUPTS

    PUSH   DS
    MOV    AX,CODESG              ; INIT DS FOR TABLE TRANSFER
    MOV    DS,AX
    MOV    CX,IBM_TBL_SIZ         ; INIT COUNT FOR TRANSFER
    MOV    SI,OFFSET IBM_VEC_TBL  ; INIT SOURCE POINTER
    MOV    DI,OFFSET IBM_INT      ; INIT DESTINATION POINTER
    CLD                           ; INIT DIRECTION
    REP    MOVSW 

; INITIALIZE HARD INTERRUPTS

    MOV    CX,MPX_TBL_SIZ         ; INIT COUNT FOR TRANSFER
    MOV    SI,OFFSET MPX_VEC_TBL  ; INIT SOURCE POINTER
    MOV    DI,OFFSET MPX_INT      ; INIT DESTINATION POINTER
    CLD                           ; INIT DIRECTION
    REP    MOVSW 
    POP    DS                     ; RESTORE NORMAL DS

;-----------------------------
; READ SWITCHES AND SET PARAMS
;-----------------------------
	MOV	DX,P8255_A		; FIRST READ IN DIP SWITCHES
	IN	AL,DX
	AND	AL,3		; ONLY KEEP NUMBER OF DISK DRIVES
	MOV	CL,6
	SHL	AL,CL			; BITS 7&8 = DISK DRIVES
	MOV	AH,AL			; SAVE DISK INFORMATION
	IN	AL,DX		; READ SWITCHES AGAIN
	TEST	AL,010H		; TEST FOR TERMINAL OR KBD/MON
	JNE	TERM_MODE
;
	SHR	AL,1		; MOVE INTO CORRECT POSITIONS
	AND	AL,030H		; ONLY KEEP MONITOR INFORMATION
	JMP	SHORT	SET_EQUIP_FLAG
;
TERM_MODE:
	XOR	AL,AL		; NO MONITORS ATTACHED
;
SET_EQUIP_FLAG:
	OR	AL,AH		; GET NUMBER OF DISK DRIVES
	OR	AL,0DH		; DISK ATTACHED, 64K MEMORY MIN
	MOV	AH,044H		; 2 SERIAL PORTS, 1 PRINTER PORT
	MOV	EQUIP_FLAG,AX	; SAVE THE RESULTS
;

; INITIALIZE 8251 CONSOLE PORT

	MOV	BX,EQUIP_FLAG	; TERMINAL MODE?
	AND	BL,030H
	JE	GET_BAUD_RATE
;
	MOV	AL,0E0H		; ASSUME 9600 BAUD
	JMP	SHORT INIT_TERM
;
GET_BAUD_RATE:
	MOV	DX,P8255_A
	IN	AL,DX
   AND    AL,0E0H                ; BAUD RATE BITS ONLY
;
INIT_TERM:
	OR	AL,2		; 7 DATA BITS
   SUB    DX,DX                  ; CONSOLE PORT
	XOR	AH,AH		; INITIALIZE FUNCTION CODE
   INT    14H                    ; INIT CALL TO SERIAL PORT I/O
;
	MOV	DL,1
	XOR	AH,AH
	MOV	AL,082H
	INT	014H		; AUX PORT = 1200,N,1,7
;
; INITIALIZE FLGS

    MOV    DISK_INT_FLG,0         ; NO DISK INT VALID YET

;*
;*	INITIALIZE THE MONITOR ADAPTER AND MPX-16 SERIAL KEYBOARD ADAPTER.
;*
;
;	CHECK FOR SERIAL TERMINAL OR KEYBOARD/MONITOR CONFIGURATION
;	IF TERMINAL, DON'T DO THE FOLLOWING INITIALIZATIONS.
;
	MOV	AX,EQUIP_FLAG	; CHECK FOR TERMINAL OR KBD/MON MODE
	AND	AL,030H
	JE	PRINT_LDR_MESSAGE	; IF 0 THEN IN TERMINAL MODE
;
	CMP	AL,030H		; B&W MONITOR?
	JNE	CK_RES
	XOR	AL,AL		; 40X25 B$W
	JMP	INIT_MON	; 0 IN AL = 40X25 B$W
CK_RES:	CMP	AL,020H		; MED OR HI-RES?
	JE	HI_RES
	MOV	AL,1		; 40X25 B$W
	JMP	INIT_MON
HI_RES:	MOV	AL,3		; 80X25 COLOR
INIT_MON:
	XOR	AH,AH
	INT	010H
;
	CALL	KB_INIT		; INITIALIZE KEYBOARD ADAPTER
;
; PRINT A LOADER MESSAGE
;
PRINT_LDR_MESSAGE:
	XOR	AL,AL
	CALL	PRINT_MESSAGE
;
;-----------------------------
; BEGIN BOOTSTRAP
;-----------------------------

    INT    19H                    ; XFER TO BOOT ROUTINE

;-----------------------------
; ROM BIOS INTERRUPT ACCESSIBLE FUNCTIONS
;-----------------------------

IBM_CASSETTE:                     ; IBM COMPAT CASSETTE I/O  (INT 15)
IBM_BASIC:                        ; IBM ROM BASIC ENTRY      (INT 18)
IBM_KB_BREAK:                     ; IBM COMPAT KB BREAK ADDR (INT 1B)
IBM_TIME_BREAK:                   ; IBM COMPAT TIMER BRK ADR (INT 1C)
IBM_VID_EXT:                      ; IBM COMPAT VEDEO EXTENS  (INT 1F)
;
SIO1_IN_INT:                      ; CONSOLE IN            INT 48
SIO2_IN_INT:                      ; AUX IN                INT 49
SIO1_OUT_INT:                     ; CONSOLE OUT           INT 4A
SIO2_OUT_INT:                     ; AUX OUT               INT 4B
PIO_OUT_INT:                      ; PRINTER OUT           INT 4C
;
    IRET

NMI_INT    PROC    NEAR
    PUSH   AX                     ; SAVE ALL DATA REGS
    PUSH   DX
    MOV    DX,P8255_C             ; DISABLE PARITY CHECK
    MOV    AL,0F0H
    OUT    DX,AL                  ; PARITY DISABLED SO CONTINUE
    POP    DX                     ; RESTORE DATA REGS
    POP    AX
    IRET
NMI_INT        ENDP

REAL_TIME       PROC     FAR
        STI
        PUSH    DS
        PUSH    DX
        PUSH    AX
        MOV     AX,DATA
        MOV     DS,AX             ; LOAD DATA SEGMENT
        DEC     MOTOR_CNT         ; TIMEOUT FOR DRIVE MOTORS
        JNZ     RT1
        XOR     AX,AX
        MOV     MOTOR_STAT,AL     ; MOTOR ON BITS OFF
        OUT     MOTOR_PORT,AL     ; ACTUALLY TURN OFF MOTORS
RT1:    INC     TIMER_LSW
        JNZ     RT2
        INC     TIMER_MSW
        JNZ     RT2
        MOV     TIMER_FLG,1	  ; FLAG TIMER OVERFLOW
RT2:	INT	01CH		; USER TIMER BREAK
        MOV     DX,P8259_M0
        MOV     AL,060H
        OUT     DX,AL             ; RESET MASTER 8259
        POP     AX
        POP     DX
        POP     DS
        IRET
REAL_TIME       ENDP

;--INT 19 ------------------------
;BOOT STRAP LOADER
;    TRACK 0, SECTOR 1 OF DRIVE 0 IS READ INTO MEMORY AT
;    0000:7C00H AND CONTROL IS TRANSFERRED THERE.
;
;    IF A SUCCESSFUL READ IS NOT
;    ACCOMPLISHED, THEN CONTROL IS TRANSFERRED TO THE MONITOR
;    PROM AT FC00:0000H.
;
;---------------------------------

BOOT_STRAP      PROC      FAR     ; IBM COMPAT BOOTSTRAP     (INT 19)

    STI
	XOR	SI,SI		; RETRY COUNTER
INIT_LOOP:
	XOR	AH,AH		; INIT FUNCTION
    INT    13H                    ; INIT FDC
    JNC     READ_SEC
	INC	SI		; FOUR RETRIES?
	CMP	SI,4
	JAE	BS1
	JMP	SHORT	INIT_LOOP
READ_SEC:
	XOR	SI,SI		; RETRY COUNTER
READ_LOOP:
    MOV    AL,1
    MOV    AH,2
    MOV    BX,0
    MOV    ES,BX
    MOV    BX,7C00H
    MOV    CX,1
    MOV    DX,0
    INT    13H                    ; READ IN SECTOR 0,TRACK 0
    JNC     GOTO_DOS
	INC	SI		; FOUR RETRIES?
	CMP	SI,4
	JAE	BS1
	JMP	SHORT READ_LOOP
;
GOTO_DOS:
    MOV    AX,0
    PUSH   AX
    MOV    AX,7C00H
    PUSH   AX
    RET

BS1: MOV	AL,2
	CALL	PRINT_MESSAGE
    MOV    AX,0FC00H              ; ERROR, GO TO MONITOR
    PUSH   AX
    MOV    AX,0
    PUSH   AX
    RET

BOOT_STRAP    ENDP

;--INT 1A--
;
;  TIME OF DAY
;
;  AH=0 MEANS READ, AH=1 MEANS WRITE
;
;  CURRENT CLOCK READS IN 50MS UNITS (NON IBM)
;
;----------

TIME_DAY       PROC    FAR
        STI
        PUSH  DS
        PUSH  AX
        MOV   AX,DATA
        MOV   DS,AX
        POP   AX
        OR    AH,AH
        JNZ    TD1
        CLI
        MOV   AL,TIMER_FLG
        MOV   TIMER_FLG,0
        MOV   CX,TIMER_MSW
        MOV   DX,TIMER_LSW
        JMP   SHORT TD2
TD1:    CLI
        MOV   TIMER_FLG,0
        MOV   TIMER_MSW,CX
        MOV   TIMER_LSW,DX
TD2:    POP   DS
        IRET
TIME_DAY      ENDP

;--INT 11--
;
;  EQUIPMENT FLGS
;
;----------

IBM_EQUIP     PROC       FAR
        STI
	PUSH	DS
	MOV	AX,DATA
	MOV	DS,AX
        MOV      AX,EQUIP_FLAG
	POP	DS
        IRET
IBM_EQUIP     ENDP

;--INT 12--
;
;  MEMORY CALL
;
;----------

IBM_MEMORY    PROC        FAR
        STI
        PUSH    DS
        MOV     AX,DATA
        MOV     DS,AX
        MOV     AX,MEM_SIZ
        POP     DS
        IRET
IBM_MEMORY    ENDP

;--INT 14-------------------------
;
;  SERIAL I/O
;
;  BYTE I/O VIA BOTH CONSOLE AND AUX 8251S
;
;  (DX) = 0/CONSOLE, 1/AUX
;
;  (AH) = 0/INIT CHIP
;
;    (AL) = INIT PARAMETERS
;
;    7   6   5       4  3         2          1  0
;    BAUD RATE     -PARITY--   STOPBITS   CHAR LENGTH
;
;    000 - 110     X0 - NONE    0 - 1     10 - 7 BITS
;    001 - 150     01 - ODD     1 - 2     11 - 8 BITS
;    010 - 300     11 - EVEN
;    011 - 600
;    100 - 1200
;    101 - 2400
;    110 - 4800
;    111 - 9600
;
;  (AH) = 1/SEND CHAR
;
;    (AL) = CHAR TO BE SENT
;
;  (AH) = 2/RECV CHAR
;
;    (AL) = WILL CONTAIN CHAR RECVD
;
;  (AH) = 3/GET STATUS
;
;    (AL) = WILL CONTAIN STATUS BITS
;
;---------------------------------

SER_IO        PROC    FAR

    STI                           ; ENABLE INTERRUPTS
    PUSH   DS
    PUSH   BX                     ; SAVE REGS
    PUSH   CX
    PUSH   DX
    MOV    CX,DATA
    MOV    DS,CX
    MOV    CX,AX                  ; SAVE PARAMETERS
    OR     DX,DX                  ; GET PORT SELECTION
    MOV    DX,P8251_CON           ; DEFAULT TO CONSOLE
    JZ     S1
    MOV    DX,P8251_AUX           ; USE AUX PORT
S1: OR     AH,AH                  ; TEST WHICH FUNCTION
    JZ     INIT_SER_IO
    DEC    AH
    JZ     SEND_SER_IO
    DEC    AH
    JZ     RECV_SER_IO
    DEC    AH
    JZ     STAT_SER_IO

SER_IO_RET:

    POP    DX                     ; RESTORE REGS
    POP    CX
    POP    BX
    POP    DS
    IRET

INIT_SER_IO:

    SUB    AL,AL                  ; RESET 8251
    OUT    DX,AL
    OUT    DX,AL
    OUT    DX,AL
    MOV    AL,40H
    OUT    DX,AL
    MOV    AL,CL                  ; CONSTRUCT MODE BYTE
    AND    AL,18H                 ; ISOLATE PARITY BITS
    SHL    AL,1                   ; SHIFT INTO PLACE
    OR     AL,42H                 ; SET 16X BAUD FACTOR, 1 STOP BIT
    TEST   CL,4                   ; 2 STOP BITS?
    JZ     S3                     ; NO
    OR     AL,80H                 ; SET 2 STOP BITS
S3: MOV    AH,CL                  ; SAVE INTERMEDIATE RESULTS
    AND    AH,3                   ; ISOLATE WORD LEN BITS
    SHL    AH,1                   ; SHIFT INTO PLACE
    SHL    AH,1
    OR     AL,AH                  ; XFR BITS TO MODE BYTE
    OUT    DX,AL                  ; OUTPUT MODE BYTE
    MOV    AL,37                  ; ENABLE COMMUNICATION
    OUT    DX,AL
    MOV    AL,CL                  ; GET BAUD RATE INFO
    SHR    AL,1                   ; MAKE WORD INDEX
    SHR    AL,1
    SHR    AL,1
    SHR    AL,1
    AND    AX,0EH                 ; CLEAN UP INDEX
    MOV    BX,OFFSET BAUD_RATE    ; GET TABLE ADDR
    ADD    BX,AX                  ; POINT TO TABLE VALUE
    MOV    BX,CS:[BX]             ; GET COUNTER VALUE
    MOV    CX,P8253_0             ; LOAD DEFAULT TIMER CHANNEL (CONSOLE)
    MOV    AL,36H                 ; LOAD DEFAULT COMMAND
    CMP    DX,P8251_CON           ; CONSOLE?
    JE     S4                     ; YES
    MOV    CX,P8253_1             ; LOAD AUX TIMER CHANNEL
    MOV    AL,76H                 ; LOAD AUX COMMAND
S4: MOV    DX,P8253_CMD           ; COMMAND PORT
    OUT    DX,AL                  ; SEND TO 8253
    MOV    DX,CX                  ; GET 8253 CHANNEL ADDR
    MOV    AL,BL                  ; GET LSB
    OUT    DX,AL
    MOV    AL,BH                  ; GET MSB
    OUT    DX,AL
    CMP    DX,P8253_0             ; CONSOLE ADDR?
    MOV    DX,P8251_CON           ; RESTORE CON ADDR
    JE     STAT_SER_IO            ; DEFAULT WAS CORRECT
    MOV    DX,P8251_AUX           ; RESTORE AUX ADDR

STAT_SER_IO:

    IN     AL,DX                  ; GET STATUS
    JMP    SER_IO_RET             ; RETURN

SEND_SER_IO:

    MOV    AH,AL                  ; SAVE CHAR
S5: IN     AL,DX                  ; GET STATUS
    TEST   AL,1                   ; READY?
    JZ     S5                     ; NOT YET
    MOV    AL,AH                  ; GET CHAR
    DEC    DX                     ; POINT TO DATA REG
    OUT    DX,AL                  ; SEND CHAR
    JMP    SER_IO_RET             ; RETURN

RECV_SER_IO:

    IN     AL,DX                  ; GET STATUS
    TEST   AL,2                   ; CHAR READY?
    JZ     RECV_SER_IO            ; NO
    DEC    DX                     ; POINT TO DATA REG
    IN     AL,DX                  ; GET CHAR
    JMP    SER_IO_RET
SER_IO     ENDP

BAUD_RATE    LABEL    WORD

        DW    698                 ; 110
        DW    512                 ; 150
        DW    256                 ; 300
        DW    128                 ; 600
        DW    64                  ; 1200
        DW    32                  ; 2400
        DW    16                  ; 4800
        DW    8                   ; 9600


;--INT 17 ------------------------
;
;  PRINTER I/O
;
;  (AH) = 0/SEND CHAR
;    (AL) CONTAINS THE CHAR TO BE SENT
;
;  (AH) = 1/SEND INIT TO PRINTER
;
;  (AH) = 2/GET STATUS
;    (AH) WILL CONTAIN THE STATUS
;
;----------------------------------

PRINT_IO    PROC    FAR

    STI                           ; ENABLE INTERRUPTS
	PUSH	BX
    PUSH   CX
    PUSH   DX
    MOV    DX,P8255_C             ; STATUS LINES
    OR     AH,AH                  ; TEST WHICH FUNCTION
    JZ     SEND_PRINT_IO
    DEC    AH
    JZ     INIT_PRINT_IO
    DEC    AH
    JZ     STAT_PRINT_IO

PRINT_IO_RET:

    POP    DX                     ; RESTORE REGS
    POP    CX
	POP	BX
    IRET

INIT_PRINT_IO:

    XCHG   AH,AL                  ; SAVE AL
    MOV    AL,0EH                 ; SEND INPUT PRIME* LOW (RESET)
    INC    DX                     ; USE CONTROL PORT SET/RESET FEATURE
    OUT    DX,AL                  ; RESET PRINTER
    MOV    CX,1000H               ; DELAY
P1: LOOP   P1
    MOV    AL,0FH                 ; INPUT PRIME* HIGH
    OUT    DX,AL
    DEC    DX                     ; POINT TO STATUS LINES
    XCHG   AH,AL                  ; RESTORE AL
    JMP    STAT_PRINT_IO          ; GO GET STATUS

SEND_PRINT_IO:

    MOV    AH,AL                  ; SAVE CHAR
    DEC    DX                     ; DATA PORT B
    OUT    DX,AL                  ; SETUP DATA LINES
    INC    DX                     ; STATUS PORT C
	MOV	BL,11		; NUMBER OF TIMEOUTS
    SUB    CX,CX                  ; TIMEOUT VALUE
P2: IN     AL,DX                  ; GET STATUS
    TEST   AL,8                   ; BUSY?
    JNZ    P3                     ; NO
    LOOP   P2                     ; TRY AGAIN
	DEC	BL
	JNZ	P2		; TRY AGAIN
    AND    AL,0FH                 ; STATUS BITS ONLY
    OR     AL,81H                 ; TURN ON TIMEOUT BIT
    XCHG   AH,AL                  ; STATUS/CHAR IN PROPER BYTES
    JMP    PRINT_IO_RET           ; RETURN
P3: INC    DX                     ; USE CNTRL BIT SET/RESET FEATURE
    MOV    AL,0CH                 ; STROBE LOW
    OUT    DX,AL
    MOV    AL,0DH                 ; STROBE HIGH
    OUT    DX,AL
    DEC    DX                     ; POINT TO STATUS LINES

STAT_PRINT_IO:

    XCHG   AH,AL                 ; SAVE AL
    IN     AL,DX                 ; GET STATUS
    AND    AL,0FH                ; STATUS BITS ONLY
    XCHG   AH,AL                 ; RESTORE AL
    JMP    PRINT_IO_RET          ; RETURN

PRINT_IO   ENDP


;--INT 13 ------------------------
;DISK I/O
;
;  INPUT:
;       (AH)=0  RESET DISKETTE SYSTEM
;               HARD RESET TO 8272, SPECIFY COMMAND, RECAL REQD
;       (AH)=1  READ STATUS INTO (AL)
;               DISK_STAT FROM LAST XFR IS USED
;       (AH)=2  READ SECTORS INTO MEMORY
;       (AH)=3  WRITE SECTORS FROM MEMORY
;       (AH)=4  VERIFY SECTORS
;       (AH)=5  FORMAT A TRACK
;                 THE BUFFER POINTER (ES,BX) MUST
;                 POINT TO THE DESIRED ADDRESS FIELDS FOR THE
;                 TRACK.  EACH FIELD IS COMPOSED OF 4 BYTES, (C,H,R,N),WHERE
;                 C = TRACK NUMBER, H=HEAD NUMBER, R = SECTOR NUMBER, N=NUMBER
;                 OF BYTES/SECTOR (00=128, 01=256, 02=512, 03=1024,)
;                 THERE MUST BE ONE ENTRY FOR EVERY SECTOR ON THE TRACK.
;
;  REGISTERS FOR READ/WRITE/VERIFY/FORMAT
;       (DL) - DRIVE NUMBER (0-3)
;       (DH) - HEAD NUMBER (0-1)
;       (CH) - TRACK NUMBER (0-39)
;       (CL) - SECTOR NUMBER (1-8)
;       (AL) - NUMBER OF SECTORS (MAX=8)
;       (ES:BX) - ADDRESS OF BUFFER (NOT REQUIRED FOR VERIFY)
;
; DATA VARIABLE -- DISK_POINTER
;      DOUBLE WORD POINTER TO THE CURRENT SET OF DISKETTE PARAMETERS
;
; OUTPUT
;      AH = STATUS OF OPERATION
;      CF = 0  SUCCESSFUL OPERATION (AH=0 ON RETURN)
;      CF = 1  FAILED OPERATION (AH HAS ERROR CODE)
;      FOR READ/WRITE/VERIFY
;              DS,BX,DX,CX PRESERVED
;              AL = NUMBER OF SECTORS ACTUALLY READ
;                (AL MAY NOT BE CORRECT IF TIME OUT OCCURS)
;------------------------------------------

DISK_IO         PROC        FAR
        STI
	PUSH	DS
        PUSH    DI
        PUSH    SI
        PUSH    DX
        PUSH    CX
        PUSH    BX
	MOV	DI,DATA				; USE CORRECT DS
	MOV	DS,DI
        OR      AH,AH                       ; INIT FUNC
        JNZ     DD1
        CALL    DISK_RESET
        JMP     SHORT DISK_IO_RET
DD1:    DEC     AH                          ; GET STATUS FUNC
        JNZ     DD2
        MOV     AL,DISK_STAT
        JMP     SHORT DISK_IO_RET
DD2:    MOV     NUMSEC,AL                   ; SAVE FOR CMDS
        DEC     AH                          ; READ FUNC
        JNZ     DD3
        CALL    DISK_READ
        JMP     SHORT DISK_IO_RET
DD3:    DEC     AH                          ; WRITE FUNC
        JNZ     DD4
        CALL    DISK_WRITE
        JMP     SHORT DISK_IO_RET
DD4:    DEC     AH                          ; VERIFY FUNC
        JNZ     DD5
        CALL    DISK_VERIFY
        JMP     SHORT DISK_IO_RET
DD5:    DEC     AH                          ; FORMAT FUNC
        JNZ     DD6
        CALL    DISK_FORMAT
        JMP     SHORT DISK_IO_RET
DD6:    MOV     DISK_STAT,BAD_CMD           ; FLAG CMD ERROR

DISK_IO_RET:
        MOV     AH,DISK_STAT
        OR      AH,AH                       ; ANY ERRORS ?
        JZ      DD7                         ; NO
        STC
        JMP     SHORT DD8
DD7:    CLC
DD8:    POP     BX
        POP     CX
        POP     DX
        POP     SI
        POP     DI
	POP	DS
        RET     2                           ; USE RET 2 /NOT IRET
DISK_IO         ENDP

DISK_RESET      PROC    NEAR
        CLI
	MOV	DISK_INT_FLG,0			; CLEAR FLAG FIRST
        MOV     AL,MOTOR_STAT               ; KEEP MOTORS RUNNING
        OR      AL,8                        ; FDC RESET
        OUT     MOTOR_PORT,AL
        MOV     DISK_STAT,0                 ; CLEAR STATUS
        AND     AL,7                        ; FDC RESET BACK TO NORMAL
        OUT     MOTOR_PORT,AL
        CALL    WAIT_INT
        JC      DS1
        MOV     AH,08H                      ; SENSE INT STAT CMD
        CALL    FDC_CMD                     ; SEND TO FDC
        CALL    FDC_RSLTS                   ; GET STATUS
        JC      DS1
        MOV     RECAL_FLG,0                 ; ALL DRIVES NEED RECAL
        MOV     AH,03H                      ; SPECIFY CMD TO FDC
        CALL    FDC_CMD
        PUSH    ES
	MOV	SI,ABS0
	MOV	ES,SI
        LES     SI,DISK_POINTER             ; GET TABLE ADDR
        MOV     AH,ES:[SI]                  ; 1ST SPECIFY BYTE
        CALL    FDC_CMD
        MOV     AH,ES:[SI+1]                ; 2ND SPECIFY BYTE
        CALL    FDC_CMD
        POP     ES
        RET

DS1:    OR      DISK_STAT,TIME_OUT          ; FLAG TIMEOUT ERROR
        STC
        RET
DISK_RESET      ENDP

DISK_READ       PROC       NEAR
        MOV     AL,046H                     ; DMA READ
        CALL    DISK_DMA                    ; SETUP XFR
        JC      DR10                        ; BOUNDARY ERROR
        CALL    DISK_MOTOR                  ; MOTOR ON
        MOV     AH,066H                     ; READ COMMAND
        CALL    DISK_XFR                    ; DO IT
        RET

DR10:   OR      DISK_STAT,DMA_BOUNDARY      ; FLAG ERROR
        MOV     AL,0                        ; NO SECTORS XFRD
        RET
DISK_READ       ENDP

DISK_WRITE      PROC       NEAR
        MOV     AL,04AH                     ; DMA WRITE
        CALL    DISK_DMA                    ; SETUP XFR
        JC      DW10                        ; BOUNDARY ERROR
        CALL    DISK_MOTOR                  ; MOTOR ON
        MOV     AH,045H                     ; WRITE COMMAND
        CALL    DISK_XFR                    ; DO IT
        RET

DW10:   OR      DISK_STAT,DMA_BOUNDARY      ; FLAG ERROR
        MOV     AL,0                        ; NO SECTORS XFRD
        RET
DISK_WRITE      ENDP

DISK_VERIFY     PROC       NEAR
        MOV     AL,042H                     ; DMA VERIFY
        CALL    DISK_DMA                    ; SETUP XFR
        JC      DV10                        ; BOUNDARY ERROR
        CALL    DISK_MOTOR                  ; MOTOR ON
        MOV     AH,066H                     ; READ COMMAND
        CALL    DISK_XFR                    ; DO IT
        RET

DV10:   OR      DISK_STAT,DMA_BOUNDARY      ; FLAG ERROR
        MOV     AL,0                        ; NO SECTORS XFRD
        RET
DISK_VERIFY     ENDP

DISK_FORMAT     PROC       NEAR
        MOV     AL,04AH                     ; DMA WRITE
        CALL    DISK_DMA                    ; SETUP XFR
        JC      DF10                        ; BOUNDARY ERROR
        CALL    DISK_MOTOR                  ; MOTOR ON
        MOV     AH,04DH                     ; FORMAT COMMAND
        CALL    DISK_XFR                    ; DO IT
        RET

DF10:   OR      DISK_STAT,DMA_BOUNDARY      ; FLAG ERROR
        MOV     AL,0                        ; NO SECTORS XFRD
        RET
DISK_FORMAT     ENDP

DISK_DMA        PROC       NEAR
        PUSH    CX
        OUT     P8237+12,AL                 ; RESET LO/HO TOGGLE
        OUT     P8237+11,AL                 ; OUTPUT MODE BYTE
        MOV     AX,ES                       ; GET SEGMENT
        ROL     AX,1                        ; ISOLATE 4 UPPER BITS
        ROL     AX,1
        ROL     AX,1
        ROL     AX,1
        MOV     CL,AL                       ; SAVE
        AND     AL,0F0H                     ; PAGE OFFSET VALUE ONLY
        ADD     AX,BX
        JNC     DA1                         ; NO CARRY INTO PAGE VALUE
        INC     CL
DA1:    PUSH    AX                          ; SAVE ADDR
        OUT     P8237+4,AL                  ; LO ADDR
        MOV     AL,AH
        OUT     P8237+4,AL                  ; HO ADDR
        MOV     AL,CL
        AND     AL,0FH                      ; PAGE VALUE ONLY
        OUT     PAGE_PORT,AL                ; SET UP DMA PAGE REGISTER
        SUB     AX,AX
        MOV     AH,NUMSEC                   ; # OF SECTORS
        SHR     AX,1                        ; SECTORS*128
        PUSH    ES
	MOV	SI,ABS0
	MOV	ES,SI
        LES     SI,DISK_POINTER
        MOV     CL,ES:[SI+3]                ; GET BYTES/SECTOR CODE
        POP     ES
        SHL     AX,CL                       ; MULTIPLY BY CODE
        DEC     AX                          ; FOR DMA CHIP ARITHMETIC
        PUSH    AX                          ; STORE COUNT
        OUT     P8237+5,AL                  ; OUTPUT LO COUNT
        MOV     AL,AH
        OUT     P8237+5,AL                  ; OUTPUT HO COUNT
        POP     AX                          ; GET COUNT
        POP     CX                          ; GET ADDR
        ADD     AX,CX                       ; SET CF IF OVERFLOW
        MOV     AL,2
        OUT     P8237+10,AL                 ; INIT CHANNEL
        POP     CX
        RET
DISK_DMA        ENDP

DISK_MOTOR      PROC       NEAR
        PUSH    CX
;
;	FIRST RESET MOTOR_ON COUNTER SO THE REAL TIME INTERRUPT DOESN'T
;	TURN THE MOTORS OFF
;
	PUSH	ES
	MOV	SI,ABS0
	MOV	ES,SI
	LES	SI,DISK_POINTER
	MOV	CL,ES:[SI+2]
	POP	ES
	MOV	MOTOR_CNT,CL
;
;	NOW TURN ON THE MOTORS IF NOT ALREADY ON
;
        MOV     AL,DL                       ; GET DRIVE #
        INC     AL
        INC     AL
        CMP     AL,3                        ; DRIVE B ?
        JG      DM1                         ; DRIVE C/D
        JL      DM2                         ; DRIVE A
        MOV     AL,3                        ; LOAD MASK
        JMP     SHORT DM3
DM1:    MOV     AL,4
        JMP     SHORT DM3
DM2:    MOV     AL,3                        ; LOAD MASK
DM3:    TEST    AL,MOTOR_STAT               ; MOTOR ON ?
        JNZ     DM6                         ; YES
        MOV     AH,MOTOR_STAT               ; FLAG ON NOW
        OR      AL,AH
	MOV	MOTOR_STAT,AL
        OUT     MOTOR_PORT,AL
        PUSH    ES
	MOV	SI,ABS0
	MOV	ES,SI
        LES     SI,DISK_POINTER
        MOV     AH,ES:[SI+10]               ; GET WAIT TIME
        POP     ES
DM4:    OR      AH,AH                       ; NO MORE WAIT ?
        JZ      DM6
        XOR     CX,CX
DM5:    LOOP    DM5
        DEC     AH                          ; COUNT EACH 1/8 SEC
        JMP     DM4
DM6:        POP     CX
        RET
DISK_MOTOR      ENDP

DISK_XFR        PROC      NEAR
        PUSH    AX                          ; SAVE COMMAND
        PUSH    CX
        MOV     AL,1                        ; CREATE DRIVE MASK
        MOV     CL,DL                       ; GET DRIVE #
        ROL     AL,CL
        POP     CX
        TEST    AL,RECAL_FLG                ; RECALIBRATE ?
        JNZ     DX1B                        ; NOT NEEDED
        OR      RECAL_FLG,AL                ; SET FLAG
        MOV     AH,07H                      ; RECALIBRATE
        CALL    FDC_CMD
        MOV     AH,DL
        CALL    FDC_CMD
        CALL    WAIT_INT
        JNC     DX1A
DX1:    OR      DISK_STAT,TIME_OUT          ; FLAG ERROR
        POP     AX                          ; CLEAN UP STACK
        RET
DX1A:   MOV     AH,08H                      ; SENSE INT STATUS
        CALL    FDC_CMD
        CALL    FDC_RSLTS
        JC      DX1                         ; ERROR
DX1B:   MOV     AH,0FH                      ; SEEK COMMAND
        CALL    FDC_CMD
        MOV     AH,DL                       ; DRIVE #
        CALL    FDC_CMD
        MOV     AH,CH                       ; TRACK #
        CALL    FDC_CMD
        CALL    WAIT_INT
        JC      DX1                         ; TIMEOUT ERROR
        MOV     AH,08H                      ; SENSE INT STAT
        CALL    FDC_CMD
        CALL    FDC_RSLTS
        JC      DX1                         ; ERROR
        PUSH    ES
	MOV	SI,ABS0
	MOV	ES,SI
        LES     SI,DISK_POINTER
        MOV     AH,ES:[SI+9]                ; GET HEAD SETTLE TIME
        POP     ES
        PUSH    CX
DX2:    MOV     CX,550                      ; VALUE FOR 1 MS LOOP
        OR      AH,AH
        JZ      DX3A                        ; DONE
DX3:    LOOP    DX3
        DEC     AH                          ; COUNT MS
        JMP     SHORT DX2
DX3A:   POP     CX
        POP     AX                          ; GET COMMAND
        CMP     AH,04DH                     ; FORMAT ?
        JE      DX4
        CALL    FDC_CMD                     ; SEND CMD BYTE
        MOV     AH,DH                       ; GET HEAD #
        SAL     AH,1
        SAL     AH,1
        AND     AH,4                        ; ISOLATE HEAD BIT
        AND     DL,03H                      ; ISOLATE DRIVE #
        OR      AH,DL                       ; MAKE HEAD/DRIVE #
        CALL    FDC_CMD
        MOV     AH,CH                       ; GET CYLINDER #
        CALL    FDC_CMD
        MOV     AH,DH                       ; GET HEAD #
        CALL    FDC_CMD
        MOV     AH,CL                       ; GET SECTOR #
        CALL    FDC_CMD
        PUSH    ES
	MOV	SI,ABS0
	MOV	ES,SI
        LES     SI,DISK_POINTER
        MOV     AH,ES:[SI+3]                ; GET BYTES/SECTOR
        CALL    FDC_CMD
        MOV     AH,ES:[SI+4]                ; GET # OF SECTORS
        CALL    FDC_CMD
        MOV     AH,ES:[SI+5]                ; GET GAP LEN
        CALL    FDC_CMD
        MOV     AH,ES:[SI+6]                ; GET DTL
        CALL    FDC_CMD
        POP     ES
        JMP     DX5
DX4:    CALL    FDC_CMD                     ; SEND COMMAND BYTE
        MOV     AH,DH                       ; GET HEAD #
        SAL     AH,1
        SAL     AH,1
        AND     AH,4                        ; ISOLATE HEAD BIT
        AND     DL,03H                      ; ISOLATE DRIVE #
        OR      AH,DL                       ; MAKE HEAD/DRIVE #
        CALL    FDC_CMD
        PUSH    ES
	MOV	SI,ABS0
	MOV	ES,SI
        LES     SI,DISK_POINTER
        MOV     AH,ES:[SI+3]                ; GET BYTES/SECTOR
        CALL    FDC_CMD
        MOV     AH,ES:[SI+4]                ; GET # OF SECTORS
        CALL    FDC_CMD
        MOV     AH,ES:[SI+7]                ; GET FORMAT GAP LEN
        CALL    FDC_CMD
        MOV     AH,ES:[SI+8]                ; GET FILLER BYTE
        CALL    FDC_CMD
        POP     ES
DX5:    CALL    WAIT_INT
        JNC     DX6                         ; NO ERROR
        CALL    FDC_RSLTS                   ; FLUSH CHIP
        RET
DX6:    CALL    FDC_RSLTS                   ; GET STATUS
        JNC     DX7                         ; EVERYTHING IS OKAY
        RET
DX7:    MOV     AL,DISK_STAT_ARRAY          ; GET STATUS
        AND     AL,0C0H                     ; NORMAL ?
        JNZ     DX9                         ; ERROR, GO GET SPECIFIC

DX7A:   MOV    AL,DISK_STAT_ARRAY+3         ; GET FINAL TRACK
        CMP    AL,CH                        ; SAME AS FIRST ?
        MOV    AL,DISK_STAT_ARRAY+5         ; GET FINAL SECTOR
        JZ     DX8                          ; DON'T ADJUST
        PUSH   ES
	MOV	SI,ABS0
	MOV	ES,SI
        LES    SI,DISK_POINTER
        MOV    AL,ES:[SI+4]                 ; GET # OF SECTORS
        POP    ES
        INC    AL                           ; USE EOT+1 FOR CALCULATION
DX8:    SUB    AL,CL                        ; SUBTRACT START FROM END
        RET

DX9:    MOV    AL,DISK_STAT_ARRAY+1         ; GET STATUS 1
        TEST   AL,080H                      ; EOT ?
        MOV    AH,RECORD_NOT_FND
        JNZ    DX9A
        TEST   AL,020H                      ; CRC BAD ?
        MOV    AH,BAD_CRC
        JNZ    DX9A
        TEST   AL,010H                      ; DMA BAD ?
        MOV    AH,BAD_DMA
        JNZ    DX9A
        TEST   AL,04H                       ; RECORD NOT FOUND ?
        MOV    AH,RECORD_NOT_FND
        JNZ    DX9A
        TEST   AL,02H                       ; WRITE PROTECT ?
        MOV    AH,WRITE_PROTECT
        JNZ    DX9A
        TEST   AL,01H                       ; MISSING ADDR MARK ?
        MOV    AH,BAD_ADDR_MARK
        JNZ    DX9A
        MOV    AH,BAD_8272                  ; CATCH-ALL ERROR
DX9A:   OR     DISK_STAT,AH                 ; SET ERROR FLAG
        JMP    DX7A                         ; GO EXIT

DISK_XFR        ENDP

FDC_CMD         PROC       NEAR
        PUSH    CX
        XOR     CX,CX                       ; CLR FOR TIMEOUT
DN1:    IN      AL,P8272_STAT               ; GET STATUS
        AND     AL,0C0H
        XOR     AL,080H
        JZ      DN2                         ; OKAY
        LOOP    DN1
        OR      DISK_STAT,TIME_OUT          ; FLAG TIMEOUT
        STC
        POP     CX
        RET
DN2:    MOV     AL,AH
        OUT     P8272_DATA,AL               ; OUTPUT COMMAND BYTE
        POP     CX
        RET
FDC_CMD         ENDP

FDC_RSLTS       PROC        NEAR
        PUSH    CX
        MOV     SI,OFFSET DISK_STAT_ARRAY
        XOR     CX,CX
DR1:    IN      AL,P8272_STAT               ; GET STATUS
        AND     AL,0C0H
        XOR     AL,0C0H
        JZ      DR2                         ; OKAY
        LOOP    DR1
        OR      DISK_STAT,TIME_OUT          ; FLAG TIMEOUT
        STC
        POP     CX
        RET
DR2:    IN      AL,P8272_DATA               ; GET STATUS FROM CHIP
        MOV     [SI],AL                     ; STORE IN ARRAY
        INC     SI
        MOV     CX,010H
DR3:    LOOP    DR3                         ; KILL TIME
        IN      AL,P8272_STAT               ; GET STATUS
        AND     AL,010H                     ; FDC STILL BUSY ?
        JNZ     DR1                         ; YES        
        POP     CX
        RET
FDC_RSLTS       ENDP

WAIT_INT        PROC       NEAR
        STI
        PUSH    CX
	PUSH	AX
        XOR     CX,CX                       ; FOR TIMEOUT
	MOV	AX,CX
DW1:    CMP     DISK_INT_FLG,0FFH
        JE      DW2
        LOOP    DW1
	CMP	AL,0
	JNE	DW3
	INC	AL
	XOR	CX,CX
	JMP	SHORT DW1
DW3:    OR      DISK_STAT,TIME_OUT          ; FLAG ERROR
        STC
	JMP	SHORT DW4
DW2:    MOV     DISK_INT_FLG,0              ; CLEAR FLAG FOR NEXT TIME
DW4:	POP	AX
        POP     CX
        RET
WAIT_INT        ENDP

DISK_INT        PROC       FAR
        STI
        PUSH    AX
        PUSH    DX
        PUSH    DS
        MOV     AX,DATA
        MOV     DS,AX
        MOV     DISK_INT_FLG,0FFH           ; FLAG THAT INT OCCURRED
        MOV     DX,P8259_M0
        MOV     AL,061H
        OUT     DX,AL                       ; RESET MASTER 8259
        MOV     DX,P8259_S0
        MOV     AL,065H
        OUT     DX,AL                       ; RESET SLAVE 8259
        POP     DS
        POP     DX
        POP     AX
        IRET
DISK_INT        ENDP

;----------------------------
; INTERRUPT VECTOR TABLE
;----------------------------
IBM_VEC_TBL  EQU    $
    DW    OFFSET VID_INT_HDLR   ; IBM COMPATIBLE VIDEO I/O (INT 10H)
    DW    CODESG
    DW    OFFSET IBM_EQUIP      ; IBM COMPATIBLE EQUIP CHK (INT 11)
    DW    CODESG
    DW    OFFSET IBM_MEMORY     ; IBM COMPATIBLE MEM CHK   (INT 12)
    DW    CODESG
    DW    OFFSET DISK_IO        ; IBM COMPATIBLE DISK I/O  (INT 13)
    DW    CODESG
    DW    OFFSET SER_IO         ; IBM COMPATIBLE RS232 I/O (INT 14)
    DW    CODESG
    DW    OFFSET IBM_CASSETTE   ; IBM COMPAT CASSETTE I/O  (INT 15)
    DW    CODESG
    DW    OFFSET KB_IO          ; IBM COMPAT KEYBOARD I/O  (INT 16H)
    DW    CODESG
    DW    OFFSET PRINT_IO       ; IBM COMPAT PRINTER I/O   (INT 17)
    DW    CODESG
    DW    OFFSET IBM_BASIC      ; IBM ROM BASIC ENTRY      (INT 18)
    DW    CODESG
    DW    OFFSET BOOT_STRAP     ; IBM COMPAT BOOTSTRAP     (INT 19)
    DW    CODESG
    DW    OFFSET TIME_DAY       ; TIME OF DAY              (INT 1A)
    DW    CODESG
    DW    OFFSET IBM_KB_BREAK   ; IBM COMPAT KB BREAK ADDR (INT 1B)
    DW    CODESG
    DW    OFFSET IBM_TIME_BREAK ; IBM COMPAT TIMER BRK ADR (INT 1C)
    DW    CODESG
    DW    OFFSET VID_PARM_TABLE ; IBM COMPAT VIDEO PARAMS  (INT 1D)
    DW    CODESG
    DW    OFFSET DISK_BASE      ; IBM COMPAT DISK TABLE    (INT 1E)
    DW    CODESG
    DW    OFFSET IBM_VID_EXT    ; IBM COMPAT VIDEO EXTENS  (INT 1F)
    DW    CODESG
END_IBM_TBL    EQU    $
IBM_TBL_SIZ    EQU    (END_IBM_TBL-IBM_VEC_TBL)/2

;
; 20H-3FH RESERVED FOR MS DOS
;
;
; BEGIN MPX HARD INTERRUPTS
;
MPX_VEC_TBL    EQU    $
    DW    OFFSET REAL_TIME      ; REAL TIME INTERRUPT - INT 40H
    DW    CODESG
;
    DD    0			; INT 41H - SLAVE 8259A
    DW    OFFSET KB_INT_HDLR	; INT 042H - MPX-16 KBD ADAPTER
    DW    CODESG
    DD    05H DUP (0)		; RESERVE		INT 43H-47H
;
    DW    OFFSET SIO1_IN_INT    ; CONSOLE IN            INT 48
    DW    CODESG
    DW    OFFSET SIO2_IN_INT    ; AUX IN                INT 49
    DW    CODESG
    DW    OFFSET SIO1_OUT_INT   ; CONSOLE OUT           INT 4A
    DW    CODESG
    DW    OFFSET SIO2_OUT_INT   ; AUX OUT               INT 4B
    DW    CODESG
    DW    OFFSET PIO_OUT_INT    ; PRINTER OUT           INT 4C
    DW    CODESG
    DW    OFFSET DISK_INT       ; P8272 INTERRUPT       INT 4D
    DW    CODESG
;
    DD    02H DUP(0)            ; RESERVE INT 4EH-4FH
;
END_MPX_TBL    EQU    $
MPX_TBL_SIZ    EQU    (END_MPX_TBL-MPX_VEC_TBL)/2

;------------------------------------------------
; DISK_BASE
;------------------------------------------------

DISK_BASE    LABEL    BYTE
          DB    11001111B    ; 1ST SPECIFY BYTE
          DB    2            ; 2ND SPECIFY BYTE
          DB    025H         ; WAIT UNTIL MOTOR OFF
          DB    2            ; 512 BYTES/SECTOR
          DB    8            ; EOT
          DB    02AH         ; GAP
          DB    0FFH         ; DTL
          DB    050H         ; GAP FOR FORMAT
          DB    0F6H         ; FILL FOR FORMAT
          DB    25           ; HEAD SETTLE TIME (MILLISECONDS)
          DB    4            ; MOTOR START TIME (1/8 SECONDS)
;
START	ENDP
;
CSEG	ENDS
            END START
