;
title	MPX-16 serial keyboard adapter drivers
page	60,132
;
;**********************************************************************
;*
;*      module name - kbdvxx.a86        xx = revision number
;*
;*      target machine - the Micromint MPX-16
;*
;*      language - Intel 8088 assembly language
;*		    assembled using Microsoft MACRO assembler
;*
;*      purpose - support of an IBM Personal Computer compatible
;*                serial keyboard using the Micromint serial keyboard
;*                adapter module.
;*
;*      module origination date - 4/14/83
;*
;*      latest revision date - 8/24/83
;*
;*      IBM Personal Computer is a trademark of the International
;*      Business Machines Corporation.
;*
;**********************************************************************
;*
;*      The keyboard is interfaced via an 8255 PPI and an 8-bit
;*      shift register.  The 8255 port A lines are used to read
;*      the 8-bit data value from the shift register.  This value
;*      represents the scan code returned from the keyboard.  Bits
;*      6 and 7 of the port C lines are used to disable the keyboard
;*      clock and to clear the shift register (and interrupt request)
;*      respectively.  Bit 0 of the port C lines is tied to the
;*      interrupt request line (IRQST2) to allow for programmed I/O
;*      transfers if desired.  Bit 4 of port C is tied to the reset
;*      input of the keyboard, and allows a hardware reset
;*      under software control.
;*
KB_data_port    equ     0260h           ; 8255 port a i/o address
KB_ctrl_port    equ     0262h           ; 8255 port c i/o address:
KB_intr_port    equ     0262h           ; 8255 port c i/o address
K8255_com_port  equ     0263h           ; 8255 command port i/o address
;
;**********************************************************************
;*
;*      keyboard interrupt mask bits
;*
KB_int_on       equ     0FBh    ; used to set IRQST2 mask bit to 0
KB_int_off      equ     04h     ; used to set IRQST2 mask bit to 1
KB_eoi_port     equ     0140h   ; master 8259A eoi i/o address
KB_mask_port	equ	0141h	; master 8259A mask register address
KB_eoi          equ     062h    ; specific eoi for IRQST2
KB_data_rdy     equ     1       ; mask for port C bit 1 - IRQST2
KB_enable       equ     0C0h    ; enable the keyboard adapter
KB_rst          equ     10h     ; reset keyboard and adapter
KB_clear        equ     40h     ; clear the keyboard
;
;**********************************************************************
;*
;*      keyboard status mask bits
;*      The following equates define single bit flags in the
;*      keyboard status word which is organized as two bytes.
;*
;*
;*      The first 8 bits refer to the current shift status of the
;*      keyboard and are mainly used to keep track of the toggled
;*      shift keys (caps, scroll, num, and ins).
;*
insert_state    equ     080h    ; the insert state is now active
caps_state      equ     040h    ; the caps lock state has changed
num_state       equ     020h    ; the num lock state has changed
scroll_state    equ     010h    ; the scroll lock state has changed
hold_state      equ     8       ; keyboard is in the hold state
;*
;*      The next 8 equates correspond to the key_flag variable
;*      which is used to determine if a shift key is depressed.
;*      These bits are mainly used as mask bits for testing the
;*      shift status during processing of the scan codes.
;*
ins_shift       equ     080h    ; the insert key is depressed
cap_shift       equ     040h    ; the caps lock key is depressed
num_shift       equ     020h    ; the num lock key is depressed
scl_shift       equ     010h    ; the scroll lock key is depressed
alt_shift       equ     08h     ; alt key is depressed
ctl_shift       equ     04h     ; ctrl key is depressed
lft_shift       equ     02h     ; left shift key is depressed
rgt_shift       equ     01h     ; right shift key is depressed
;*
;*      The following are equates which correspond to the scan codes
;*      returned by the serial keyboard for various control keys.
;*      The position or function of the keys can be easily changed
;*      to suit different user requirements.  First, the following
;*      equates must be checked to see if the key to be changed is
;*      included.  If so the scan code must be changed to the new
;*      key position.  Second, the scan code(s) involved will require
;*      changing the data in the four scan code conversion tables
;*      in the data area.
;*
ctrl_key        equ     29              ; control key
left_shift_key  equ     42              ; left shift key
right_shift_key equ     54              ; right shift key
alt_key         equ     56              ; alt key
caps_lock_key   equ     58              ; caps lock key
num_lock_key    equ     69              ; number lock key
scroll_lock_key equ     70              ; scroll lock key
ins_key         equ     82              ; insert key
del_key         equ     83              ; delete key
space_key       equ     57              ; space bar
;
break_bit equ   080h            ; mask for scan code break bit
;*
;*	equates used by the keyboard queue routines
;*
qtop	equ	4
qbot	equ	6
qempty	equ	0
qstat	equ	10
;
;**********************************************************************
;*
;*      The code associated with using the serial keyboard on the
;*      Micromint MPX-16 microcomputer begins here.
;*
;**********************************************************************
;*
;*	external and public symbol definitions
;*
public	KB_init,KB_io,KB_int_hdlr,prtsc_int
;
data	segment	at	040h	; bios data segment
extrn	KB_buffer:word,KB_bufptr:word,alt_input:byte
extrn	KB_state_flag:byte,KB_shift_flag:byte,break_flag:byte
extrn	prtsc_stat:byte
data	ends
;
extrn	play_tone:far,qinit:far,qadd:far,qrem:far,print_message:far
extrn	start:far
;
cseg	segment	common
assume	cs:cseg,ds:data
;
	org	0		; start module at  FE00:0000
;
	db	'MPX-16 BIOS COPR. MICROMINT 1983'
;
	db	16 dup(0FFH)	; reserve some space
;
;*
;**********************************************************************
;*
;*      Subroutine name - KB_init
;*
;*      Purpose - to initialize the hardware components, and the data
;*                structures associated with the Micromint MPX-16 serial
;*                keyboard adapter module.
;*
;*      On entry:
;*
;*              No parameters are passed.
;*
;*      On exit:
;*
;*              No parameters are returned, no registers are destroyed.
;*
;**********************************************************************
;*
KB_init	proc far
;
        push    ax
        push    bx
        push    dx
        push    si
        push    ds
        push    es
;*
;*	Set up data segments
;*
	mov	ax,data
	mov	ds,ax
	mov	es,ax		; used by qinit routine
;*
;*      set up the 8255 PPI used to interface the serial keyboard
;*
        mov     al,091h         ; a=in, b=out, c_low=in, c_high=out
        mov     dx,K8255_com_port
        out     dx,al
;*
;*	Beep the speaker on the MPX-16 keyboard adapter.
;*
	call	beeper
;*
;*      initialize the keyboard queue
;*
	mov	si,offset KB_bufptr
        mov     bx,offset KB_buffer     ; offset of keyboard buffer
        mov     ax,32	            	; length of bios buffer
        call    qinit           ; initialize the parameter block
;*
;*      Now reset the keyboard adapter and the keyboard.
;*
        call    KB_reset        ; reset the keyboard
;*
;*      Here the keyboard hardware is assumed to be OK.  The MPX-16
;*      serial keyboard adapter is enabled.
;*
KB_ok:  mov     al,KB_enable    ; enable the keyboard
        mov     dx,KB_ctrl_port
        out     dx,al
;*
;*      KB_init exit point.
;*
        pop     es      ; restore registers
        pop     ds
        pop     si
        pop     dx
        pop     bx
        pop     ax
        ret
KB_init	endp
;*
;**********************************************************************
;*
;*      Subroutine name - KB_reset
;*
;*      Purpose - to reset the MPX-16 serial keyboard adapter hardware
;*                and to clear the keyboard interrupt mask bit.
;*
;*      On entry:
;*
;*              No parameters are passed.
;*
;*      On exit:
;*
;*		No parameters returned, AX,DX are destroyed.
;*
;**********************************************************************
;
KB_reset	proc	near
;
        xor     al,al
        mov     KB_state_flag,al        ; clear keyboard flags
        mov     KB_shift_flag,al
        mov     dx,KB_ctrl_port ; reset the keyboard and clear
        mov     al,KB_rst
        out     dx,al           ; the keyboard shift register
        cli                     ; disable interrupts temporarily
;*
;*      now clear the interrupt mask bit to enable keyboard interrupts
;*
        mov     dx,KB_mask_port ; master 8259A PIC mask register
        in      al,dx           ; read the current interrupt mask value
        and     al,KB_int_on    ; enable IRQST2 interrupts
        out     dx,al           ; write new mask back to 8259A
        sti                     ; turn interrupts back on
;
kbrex:
	clc		; indicate no hardware errors for now
        ret
KB_reset	endp
;
;**********************************************************************
;*
;*      Interrupt routine name - KB_io
;*
;*      Purpose - to provide an IBM Personal Computer compatible
;*                software interface via the 016h interrupt vector for
;*                keyboard operations.
;*
;*      On entry:
;*
;*              AH contains a function code:
;*
;*      if AH = 0, 
;*              read the next character from the input queue,
;*              return ascii character in the AL register, and
;*              return the scan code in the AH register.
;*
;*      if AH = 1,
;*              return status via the processor Z flag:
;*                      Z=1 if no scan code is available
;*                      Z=0 if a scan code is available and
;*              return next buffer entry without deleting it.
;*
;*      if AH = 2,
;*              return the shift status in the AL register.
;*
;*      On exit:
;*
;*              AL and the Z flag are changed depending on the function
;*              code, AH is destroyed.
;*
;**********************************************************************
;*
KB_io	proc	far
;
        sti             ; turn interrupts back on
        push    ds
        push    bx
        push    si
	mov	bx,data
        mov     ds,bx           ; set up correct data segment
;
        cmp     ah,0
        je      read_char       ; return next keyboard entry
;
        cmp     ah,1
        je      ret_flag        ; return status in Z flag
;
        cmp     ah,2
        je      ret_status      ; return shift status of KB_flags
	jmp	short KB_io_exit
;*
;*      AH was a 0 so an ascii character must be returned.  If
;*      no characters are in the buffer then wait for one to be
;*      entered.
;*
read_char:
;
        mov     si,offset KB_bufptr
        call    qrem    ; get next element from keyboard queue
        jc      read_char       ; buffer was empty if carry flag=1
	jmp	short KB_io_exit
;*
;*      AH was a 1 so the Z flag will be used to return status
;*      of the keyboard input buffer.  If the buffer is empty,
;*      the Z flag will be set to 1.  Otherwise the Z flag will
;*      be set to 0 and the next buffer entry will be returned
;*      in the AX register, but will not be removed from the Q.
;*
ret_flag:
;
        mov     si,offset KB_bufptr
        cmp     word ptr qstat[si],qempty       ; queue empty ?
        je      KB_status_ret   ; buffer is empty if it is a zero
;
        cli                     ; turn off interrupts
        mov     bx,qtop[si]     ; get pointer to top of queue
	mov	ax,[bx]		; get the data element
	sti			; turn interrupts back on
;
KB_status_ret:
;
	pop	si
	pop	bx
	pop	ds
	ret	2		; don't want to change flags
;*
;*      AH was a 2, so just return the shift status by passing
;*      the first byte of the KB_flags variable back in the AL
;*      register.
;*
ret_status:
;
        mov     al,KB_state_flag
        and     al,0f0h
        mov     ah,KB_shift_flag
        and     ah,0fh
        or      al,ah
;*
;*      Keyboard i/o  exit point.
;*
KB_io_exit:
;
        pop     si
        pop     bx
        pop     ds
;
        iret
;
KB_io	endp
;
;**********************************************************************
;*
;*      Interrupt routine name - KB_int_hdlr
;*
;*      Purpose - to decode scan codes generated by a serial keyboard
;*                attached to the MPX-16 serial keyboard adapter and
;*                to place both the scan code and the converted code
;*                into a keyboard input buffer implemented as a queue.
;*                Entry is via a hardware interrupt from the IRQST2
;*                line of the MPX-16 I/O channel.
;*
;*      On entry:
;*
;*              No parameters are passed.
;*
;*      On exit:
;*
;*              Either, the data codes will have been added to the input
;*              queue if it was not full, or the appropriate control
;*              function will have been executed.  All codes generated
;*              are compatible with those generated by the bios routines
;*              of the IBM Personal Computer.
;*
;**********************************************************************
;
KB_int_hdlr	proc far
;
        sti             ; enable further interrupts
        push    ax
        push    bx
        push    cx
        push    dx
        push    si
        push    di
        push    ds
        push    es
;
	mov	ax,cs
	mov	es,ax		; use for rom tables
	mov	ax,data
	mov	ds,ax		; use for bios ram area
;*
;*      Set up a pointer to the keyboard input queue parameter block.
;*
        mov     si,offset KB_bufptr
;*
;*      read in the scan code from the keyboard adapter
;*
        mov     dx,KB_data_port
        in      al,dx
        mov     ah,al           ; save the scan code in ah
        mov     al,KB_clear     ; reset the keyboard adapter
        mov     dx,KB_ctrl_port
        out     dx,al
        mov     al,KB_enable    ; renable the keyboard
        out     dx,al
        mov     al,ah           ; get back the scan code
;*
;*      check for an overrun scan code of 0ffh
;*
        cmp     al,0ffh
        jne     check_for_shift_key     ; if not check for shift code
;
	call	beeper
        jmp     KB_int_exit
;*
;*      check for one of the shifted scan codes:
;*              ctrl,left shift,right shift,alt,
;*              caps lock,num lock,scroll lock,
;*              or the ins keys.
;*
check_for_shift_key:
;
        and     al,not break_bit        ; strip off the break bit
        mov     di,offset shift_key_table
        mov     cx,sktb_length
        cld                             ; go forward through table
        repne   scasb
        mov     al,ah                   ; recover the scan code
        je      is_a_shift_key_code
        jmp     not_a_shift_code
;*
;*      process the shift key scan code
;*
is_a_shift_key_code:
;
        mov     di,cx                   ; get a mask bit for the code
        mov     dl,es:shift_mask_table[di]
        test    al,break_bit            ; check if this is a break
        jnz     shift_key_break         ; shift code
;
;**********************************************************************
;*
;*      the scan code is a shift key make code
;*
;**********************************************************************
;
shift_key_make:
;
        test    dl,0f0h         ; toggle or momemtary shift key?
        jz      mom_shift_make
;
tog_shift_make:                 ; check if already in ctrl state
;
        test    KB_shift_flag,ctl_shift
        jz      check_for_ins_key
        call    process_scan_code       ; go process the code
	jmp	short KB_int_exit
;*
;*      For the ins key, only the base case is used to toggle the
;*      insert key flags.  Since the ctrl-shifted case has already
;*      been tested, only the num_locked, alternate input, and
;*      shifted cases need to be tested for.
;*
check_for_ins_key:
;
        cmp     al,ins_key              ; scan code of 82?
        je      check_for_alt_key
;
not_ins_key:
;
        call    toggle_the_state_flag
	jmp	short KB_int_exit
;
check_for_alt_key:
;
        test    KB_shift_flag,alt_shift ; in alternate input mode?
        jz      check_if_num_locked
;
in_alt_shift_mode:
;
        call    process_scan_code       ; just process normally
	jmp	short KB_int_exit
;
check_if_num_locked:
;
        test    KB_state_flag,num_state
        jz      check_for_shifted_ins   ; still have to check if shifted
;
        test    KB_shift_flag,lft_shift+rgt_shift       ; shifted ins ?
        jz      add_k83_to_buffer
;
        call    toggle_the_state_flag           ; in base case
	jmp	short KB_int_exit
;
add_k83_to_buffer:
;
        mov     al,030h         ; set ascii to 0, scan code is in ah
        call    add_code_to_buffer
	jmp	short KB_int_exit
;
check_for_shifted_ins:
;
        test    KB_shift_flag,lft_shift+rgt_shift       ; shifted ins ?
        jnz     add_key83_to_buffer
;
        call    toggle_the_state_flag           ; in base case
	jmp	short KB_int_exit
;
add_key83_to_buffer:
;
        mov     al,030h         ; set ascii to 0, scan code is in ah
        call    add_code_to_buffer
	jmp	short KB_int_exit
;
;**********************************************************************
;*
;*      process a momentary shift key make - left and right shift
;*      keys, ctrl key, alt key.
;*
;**********************************************************************
;
mom_shift_make:         ; set the shift key depressed bit
;
        or      KB_shift_flag,dl
	jmp	short KB_int_exit	; and then return
;
;**********************************************************************
;*
;*      a "break" shift key is to be processed here
;*
;**********************************************************************
;
shift_key_break:
;
        not     dl              ; clear the key depressed flag
        and     KB_shift_flag,dl
        not     dl              ; change back to a single bit set
        test    dl,0f0h         ; toggle or momemtary shift key?
        jz      mom_shift_break
;
tog_shift_break:                ; toggle shift key break here
;
	jmp	short KB_int_exit
;
mom_shift_break:                ; momentary shift key break here
;
        cmp     al,alt_key+080h ; is this the ALT input key?
        jne     sb1
;
        xor     ah,ah           ; yes, add data to keyboard queue
        mov     al,alt_input    ; return as "ascii" code
        mov     alt_input,ah    ; and clear the alterate input buffer
        or      al,al           ; don't add if its a 0
        jz      sb1
        call    qadd            ; add the code to the input queue
        jnc     sb1             ; no carry if sucessfull
;
	call	beeper
;
sb1:	jmp	short KB_int_exit
;
;**********************************************************************
;*
;*      Here the scan code is not from one of the shift keys.
;*
;**********************************************************************
;
not_a_shift_code:
;
        call    process_scan_code
;
;**********************************************************************
;*
;*      normal keyboard interrupt handler exit point
;*
;**********************************************************************
;
KB_int_exit:
;
        mov     al,KB_eoi       ; output an eoi to the 8259A
        mov     dx,KB_eoi_port  ; interrupt controller - IRQST2
        out     dx,al
;*
;*      secondary exit point - no eoi instruction issued
;*
KB_int_exit2:
        pop     es              ; restore registers
        pop     ds
        pop     di
        pop     si
        pop     dx
        pop     cx
        pop     bx
        pop     ax
;
        iret
KB_int_hdlr	endp
;
;
;**********************************************************************
;*
;*      Subroutine name - process_scan_code
;*
;*      Purpose - to translate the scan code in the AL register from
;*                one of the four scan code conversion tables.  Return
;*                scan codes or ascii codes which are -1 are ignored;
;*                Otherwise the data word is added to the keyboard input
;*                queue.
;*
;*      On entry:
;*
;*              AL and AH both contain the scan code to be converted.
;*
;*      On exit:
;*
;*              AX and BX are destroyed, the entry will be in the input
;*              queue unless the converted code was a -1.
;*
;**********************************************************************
;
process_scan_code	proc	near
;
        test    al,080h         ; break code?
        jz      non_break_code
        ret
;
non_break_code:
;
        test    KB_state_flag,hold_state        ; in hold state?
        jz      check_alt_flag
;
in_hold_state:
;
        cmp     al,num_lock_key         ; is this the num_lock code?
        je      in1                     ; if so, ignore it
        and     KB_state_flag,not hold_state    ; clear the hold flag
in1:    ret
;
check_alt_flag:
;
        test    KB_shift_flag,alt_shift ; in alternate input mode?
        jz      check_ctrl_flag
;
;**********************************************************************
;*
;*      process an alternate input scan code
;*              note - the alt key has priority over the ctrl and
;*                     shift keys.
;*
;**********************************************************************
;
process_alt_input:
;
        test    KB_shift_flag,ctl_shift         ; system reset code?
        jz      translate_alt_code
        cmp     al,del_key                      ; CTL-ALT-DEL = reset
        jne     translate_alt_code
;
system_reset:	jmp	start		; do a power-up reset
;
translate_alt_code:
;
        mov     bx,offset alt_key_scan_code     ; translate the alt code
	xlat	cs:[bx]
        cmp     al,-1           ; is this an ignore code?
        jne     check_if_keypad_entry
;
        mov     alt_input,0             ; ignore code and clear buffer
        ret
;
check_if_keypad_entry:
;
        cmp     ah,71                   ; keypad if scan code >= 71
        jb      check_for_space_key
;
        xor     ah,ah                   ; add decimal digit to buffer
        mov     dx,ax                   ; save the translated code
        mov     al,alt_input            ; get the current value
        mov     ah,10                   ; shift left by one digit
        mul     ah
        add     ax,dx                   ; add this digit in units col.
        mov     alt_input,al            ; and save the new value
        ret
;
check_for_space_key:
;
        mov     alt_input,0             ; since code isn't in keypad
        cmp     ah,space_key            ; clear the buffer, then check
        je      add_alt_input_to_buffer ; for space key
;
        mov     ah,al                   ; not space key, use pseudo scan
        xor     al,al                   ; code and ascii null
;
add_alt_input_to_buffer:
;
        call    add_code_to_buffer
        ret
;*
;*      check if this is a ctrl key input
;*
check_ctrl_flag:
;
        test    KB_shift_flag,ctl_shift ; in ctrl shift mode?
        jz      check_for_keypad_entry
;
;**********************************************************************
;*
;*      scan code is a ctrl-shifted code
;*
;**********************************************************************
;
process_ctrl_key:
;
        cmp     ah,scroll_lock_key      ; check for break condition
        jne     check_for_pause_mode
;
break_indicated:
;
        mov     si,offset KB_bufptr     ; first clear the input
        mov     bx,offset KB_buffer     ; queue
        mov     ax,ds
        mov     es,ax
        mov     ax,32		; bios buffer size
        call    qinit
	or	break_flag,080h		; set the break flag
        int     01bh                    ; issue a break interrupt
        xor     ax,ax                   ; and output a null code
	jmp	short add_ctrl_key_to_buffer
;
check_for_pause_mode:
;
        cmp     ah,num_lock_key         ; check if a pause is to be
        jne     translate_ctrl_code     ; issued
;
pause_indicated:
;
        or      KB_state_flag,hold_state        ; set the hold flag
        mov     dx,KB_eoi_port          ; and reenable KB interrupts
        mov     al,KB_eoi
        out     dx,al
;
hold_loop:
;
        test    KB_state_flag,hold_state        ; loop until a key is
        jnz     hold_loop               ; hit (except num_lock)
        pop     ax              ; fix stack since we got here from a
        jmp     KB_int_exit2    ; subroutine call, leave without eoi
;
translate_ctrl_code:
;
        mov     bx,offset ctrl_key_scan_code
	xlat	cs:[bx]		; translate the scan code
;
        cmp     ah,55           ; print screen key is special case
        jne     check_for_key_59
;
print_screen_ctrl_key:
;
        mov     ah,al           ; set ah to pseudo scan code
        xor     al,al           ; and ascii code to null
        jmp     add_ctrl_key_to_buffer
;
check_for_key_59:
;
        cmp     ah,59           ; scan codes < 59  are ok as is
        jb      add_ctrl_key_to_buffer
;
ctrl_over58:
;
        mov     ah,al           ; convert to pseudo scan code
        xor     al,al           ; and ascii null
;
add_ctrl_key_to_buffer:
;
        call    add_code_to_buffer
        ret
;
check_for_keypad_entry:
;
        cmp     ah,71           ; if scan code < 71, not in keypad
        jb      check_shift_flags
;
;**********************************************************************
;*
;*      The keypad scan codes (codes >= 71) are processed seperately
;*      because of the num_lock and shift key interactions.
;*
;**********************************************************************
;
code_in_keypad:
;
        test    KB_state_flag,num_state         ; num_lock active?
        jz      not_num_state
;
        test    KB_shift_flag,lft_shift+rgt_shift       ; shift it?
        jz      in_num_lock1
;
base_key1:
;
        mov     bx,offset base_case_scan_code   ; convert scan code
	xlat	cs:[bx]
        cmp     ah,76           ; check for special case
        jne     add_keypad_code_to_buffer
        mov     ah,-1           ; set scan code for ignore
	jmp	short add_keypad_code_to_buffer
;
in_num_lock1:
;
        mov     bx,offset upper_case_scan_code  ; convert scan code
	xlat	cs:[bx]
	jmp	short add_keypad_code_to_buffer
;
not_num_state:
;
        test    KB_shift_flag,lft_shift+rgt_shift       ; shift it?
        jnz     in_num_lock2
;
base_key2:
;
        mov     bx,offset base_case_scan_code   ; convert scan code
	xlat	cs:[bx]
        cmp     ah,76           ; check for special case
        jne     add_keypad_code_to_buffer
        mov     ah,-1           ; set scan code for ignore
	jmp	short add_keypad_code_to_buffer
;
in_num_lock2:
;
        mov     bx,offset upper_case_scan_code  ; convert scan code
	xlat	cs:[bx]
;
add_keypad_code_to_buffer:
;
        call    add_code_to_buffer              ; add it to queue
        ret
;
check_shift_flags:
;
        test    KB_shift_flag,lft_shift+rgt_shift ; shift key active?
        jz      process_base_case_code
;
;**********************************************************************
;*
;*      Process the scan code in the shifted (uppercase) state.
;*
;**********************************************************************
;
process_shift:
;
        cmp     ah,55           ; check for print screen code
        jne     code_not_55
;
print_screen:
;
        mov     al,KB_eoi       ; clear the IRQST2 IR bit
        mov     dx,KB_eoi_port
        out     dx,al
;       int     5               ; issue print screen interrupt
;
        pop     ax      ; got in here from a call, so correct
                        ; the stack pointer by popping the old IP
;
        jmp     KB_int_exit2
        iret
;
code_not_55:
;
        mov     bx,offset upper_case_scan_code
	xlat	cs:[bx]
;
        cmp     ah,59           ; scan code from a function key?
        jb      add_shifted_code_to_buffer
;
        mov     ah,al           ; yes, use pseudo scan code and
        xor     al,al           ; output an ascii null
;
add_shifted_code_to_buffer:
;
        call    add_code_to_buffer
        ret
;
;**********************************************************************
;*
;*      The scan code here will be processed as a base case code.
;*
;**********************************************************************
;
process_base_case_code:
;
        mov     bx,offset base_case_scan_code
	xlat	cs:[bx]		; convert scan code to ascii
        call    add_code_to_buffer
        ret
;*
;*      entry point for adding the scan code and converted ascii
;*      character to the keyboard queue.
;*
add_code_to_buffer:

;
        cmp     al,-1           ; ignore code?
        jne     check_pseudo_scan
        ret
;
check_pseudo_scan:
;
        cmp     ah,-1           ; pseudo scan code?
        jne     check_for_caps_state
        ret
;*
;*      check here for the caps lock state.  If caps lock is
;*      active, alphabetic characters will be translated from
;*      upper to lower case if the shift flag is set, or from 
;*      lower to upper case if it is not set.
;*
check_for_caps_state:
;
        test    KB_state_flag,caps_state        ; in caps lock if set
        jz      add_to_queue
;
        test    KB_shift_flag,lft_shift+rgt_shift  ; shift key active?
        jz      convert_lowercase_to_uppercase
;*
;*      convert upper case to lower case if alphabetic char
;*
convert_uppercase_to_lowercase:
;
        cmp     al,'A'
        jb      add_to_queue
;
        cmp     al,'Z'
        ja      add_to_queue
;
        add     al,020h         ; convert upper to lower case
	jmp	short add_to_queue
;*
;*      convert lower case to upper case if alhpabetic char
;*
convert_lowercase_to_uppercase:
;
        cmp     al,'a'
        jb      add_to_queue
;
        cmp     al,'z'
        ja      add_to_queue
;
        sub     al,020h         ; convert lower to upper case
;*
;*      add the scan code and ascii char to the keyboard queue
;*      if there is room.  DS:SI already points to the keyboard
;*      queue parameter block.
;*
add_to_queue:
;
        call    qadd
        jnc     aq1             ; carry will be clear if added
;
	call	beeper
aq1:    ret
;
process_scan_code	endp
;
;*
;**********************************************************************
;*
;*      Subroutine name - toggle_the_flag
;*
;*      Purpose - used to toggle the keyboard state flags (insert_
;*                state, caps_state, num_state, or scroll_state)
;*                on the first "make" code for the associated keys,
;*                and to set the cooresponding key depressed flags.
;*
;*      On entry:
;*
;*              DL contains a bit-mask which defines which key is being
;*              processed (refer to the shift key equates above),
;*              AL contains the scan code for the key being processed.
;*
;*      On exit:
;*
;*              KB_shift_flag and KB_state_flag are updated,
;*              the scan code is added to the keyboard input queue if
;*              first make of the INS key and AL is destroyed.
;*
;**********************************************************************
;*
toggle_the_state_flag	proc	near
;
        test    dl,KB_shift_flag        ; key already depressed
        jz      toggle_the_flag
        ret                             ; not the first make
;
toggle_the_flag:
;
        or      KB_shift_flag,dl        ; set the depressed key bit
        xor     KB_state_flag,dl        ; and toggle the state bit
        cmp     al,ins_key              ; is this the INS key?
        je      add_insert_key_to_buffer
;
        ret                     ; not ins key, just return
;
add_insert_key_to_buffer:
;
        xor     al,al                   ; scan code already in ah
        call    add_code_to_buffer      ; output ascii null in al
        ret
;
toggle_the_state_flag	endp
;
;**********************************************************************
;*
;*	Subroutine name - beeper
;*
;*	Purpose - to output a 1000 hertz tone for 1/2 second
;*
;*	On entry:
;*
;*		No parameters are passeed
;*
;*	On exit:
;*
;*		No parameters returned, all registers preserved
;*
;**********************************************************************
;*
beeper	proc	near
;
	push	ax
	push	bx
	push	dx
	mov	ax,1000
	mov	bx,20
	xor	dl,dl		; use speaker channel 0
	mov	dh,3		; set mode for timed duration
	call	play_tone
	pop	dx
	pop	bx
	pop	ax
	ret
beeper	endp
;
;**********************************************************************
;*
;*      Interrupt routine name - prtsc_int
;*
;*      Purpose - to handle the int 05h software interrupt generated by
;*                the shift prtsc key combination.
;*
;*      Note: this routine will be implemented at a later date.
;*
;**********************************************************************
;*
prtsc_int	proc
;
        iret                    ; just return for now
;
prtsc_int	endp
	db	150 dup(0FFH)	; reserve for print screen code
;
;*
;*	The following data elements are rom tables used by the MPX-16
;*	serial keyboard routines.
;*
;
;**********************************************************************
;*
;*      table for lower case characters
;*
;**********************************************************************
;
base_case_scan_code     db      0       ; to correct table offset
                        db      01bh,'1234567890-=',8,9,'qwertyuiop[]'
                        db      0dh,-1,'asdfghjkl;',027h,60h,-1,5ch
                        db      'zxcvbnm,./',-1,'*',-1,' ',-1
;
base_case_fn_code       db      0,0,0,0,0,0,0,0,0,0
                        db      -1,-1   ; shift keys
;
base_case_keypad        db      0,0,0,02Dh,0,0,0,02Bh,0,0,0,0,0
;
;**********************************************************************
;*
;*      table for upper case characters
;*
;**********************************************************************
;
upper_case_scan_code    db      0       ; for correct table offset
                        db      01bh,'!@#$%^&*()_+',8,0,'QWERTYUIOP{}'
                        db      0dh,-1,'ASDFGHJKL:"',07eh,-1,07ch
                        db      'ZXCVBNM<>?',-1,0,-1,' ',-1
;
upper_case_fn_code      db      84,85,86,87,88,89,90,91,92,93
                        db      -1,-1   ; for shift keys
;
num_lock_keypad         db      '789-456+1230.'
;
;**********************************************************************
;*
;*      ctrl key scan code table
;*
;**********************************************************************
;
ctrl_key_scan_code      db      0       ; for correct table offset
                        db      27,-1,0,-1,-1,-1,30,-1,-1,-1,-1,31,-1
                        db      127,-1,17,23,5,18,20,25,21,9,15,16,27
                        db      29,10,-1,1,19,4,6,7,8,10,11,12,-1,-1
                        db      -1,-1,28,26,24,3,22,2,14,13,-1,-1,-1,-1
                        db      114,-1,' ',-1
;
ctrl_key_fn_code        db      94,95,96,97,98,99,100,101,102,103
                        db      -1,-1   ; for shift keys
;
ctrl_key_keypad         db      119,-1,132,-1,115,-1,116,-1,117,-1,118
                        db      -1,-1
;
;**********************************************************************
;*
;*      alt key scan code table
;*
;**********************************************************************
;
alt_key_scan_code       db      0       ; for correct table offset
                        db      -1,120,121,122,123,124,125,126,127,128
                        db      129,130,131,-1,-1,16,17,18,19,20
                        db      21,22,23,24,25,-1,-1,-1,-1,30,31
                        db      32,33,34,35,36,37,38,-1,-1,-1,-1
                        db      -1,44,45,46,47,48,49,50,-1,-1,-1
                        db      -1,-1,-1,' ',-1
;
alt_key_fn_key          db      104,105,106,107,108,109,110,111
                        db      112,113,-1,-1
;
alt_key_keypad          db      7,8,9,-1,4,5,6,-1,1,2,3,0,-1
;
;**********************************************************************
;*
;*      shift key table - scan codes of keys with shift function
;*
;**********************************************************************
;
shift_key_table         db      52h,3ah,45h,46h,38h,1dh,2ah,36h
sktb_length     equ     offset $ - offset shift_key_table
;
;**********************************************************************
;*
;*      shift key mask table
;*
;**********************************************************************
;
shift_mask_table        db      1,2,4,8,010h,020h,040h,080h
;
cseg	ends
	end
